@@ -8,8 +8,8 @@ import aiohttp
 from aiohttp import web
 import json
 
-from functools import partialmethod
 from . chat import Chat, Sender
+from . reloader import run_with_reloader
 
 __author__ = "Stepan Zastupov"
 __copyright__ = "Copyright 2015-2017 Stepan Zastupov"
@@ -21,13 +21,19 @@ RETRY_TIMEOUT = 30
 RETRY_CODES = [429, 500, 502, 503, 504]
 BOTAN_URL = "https://api.botan.io/track"
 
+# Message types to be handled by bot.handle(...)
 MESSAGE_TYPES = [
     "location", "photo", "document", "audio", "voice", "sticker", "contact",
-    "venue", "video", "game", "contact", "delete_chat_photo", "new_chat_photo",
+    "venue", "video", "game", "delete_chat_photo", "new_chat_photo",
     "delete_chat_photo", "new_chat_member", "left_chat_member",
     "new_chat_title"
 ]
 
+# Update types for
+MESSAGE_UPDATES = [
+    "message", "edited_message", "channel_post", "edited_channel_post"
+]
+
 logger = logging.getLogger("aiotg")
 
 
@@ -38,35 +44,39 @@ class Bot:
     :param int api_timeout: Timeout for long polling
     :param str botan_token: Token for http://botan.io
     :param str name: Bot name
+    :param callable json_serialize: JSON serializer function. (json.dumps by default)
+    :param callable json_deserialize: JSON deserializer function. (json.loads by default)
+    :param bool default_in_groups: Enables default callback in groups
     """
 
     _running = False
     _offset = 0
 
     def __init__(self, api_token, api_timeout=API_TIMEOUT,
-                 botan_token=None, name=None):
+                 botan_token=None, name=None,
+                 json_serialize=json.dumps, json_deserialize=json.loads,
+                 default_in_groups=False):
         self.api_token = api_token
         self.api_timeout = api_timeout
         self.botan_token = botan_token
         self.name = name
+        self.json_serialize = json_serialize
+        self.json_deserialize = json_deserialize
+        self.default_in_groups = default_in_groups
         self.webhook_url = None
-        self.session = aiohttp.ClientSession()
+        self._session = None
 
         def no_handle(mt):
             return lambda chat, msg: logger.debug("no handle for %s", mt)
 
+        # Init default handlers and callbacks
         self._handlers = {mt: no_handle(mt) for mt in MESSAGE_TYPES}
         self._commands = []
-        self._default = lambda c, m: None
-        self._inline = lambda iq: None
-        self._callback = lambda c, cq: None
-
-    def __del__(self):
-        try:
-            self.session.close()
-        except:
-            # 😶
-            pass
+        self._callbacks = []
+        self._inlines = []
+        self._default = lambda chat, message: None
+        self._default_callback = lambda chat, cq: None
+        self._default_inline = lambda iq: None
 
     async def loop(self):
         """
@@ -91,10 +101,12 @@ class Bot:
             )
             self._process_updates(updates)
 
-    def run(self):
+    def run(self, debug=False, reload=None):
         """
         Convenience method for running bots in getUpdates mode
 
+        :param bool debug: Enable debug logging and automatic reloading
+        :param bool reload: Automatically reload bot on code change
         :Example:
 
         >>> if __name__ == '__main__':
@@ -102,11 +114,31 @@ class Bot:
 
         """
         loop = asyncio.get_event_loop()
+
+        logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
+
+        if reload is None:
+            reload = debug
+
         try:
-            loop.run_until_complete(self.loop())
+            if reload:
+                loop.run_until_complete(
+                    run_with_reloader(loop, self.loop(), self.stop))
+
+            else:
+                loop.run_until_complete(self.loop())
+
+        # User cancels
         except KeyboardInterrupt:
+            logger.debug("User cancelled")
             self.stop()
 
+        # Stop loop
+        finally:
+            logger.debug("Closing loop")
+            loop.stop()
+            loop.close()
+
     def run_webhook(self, webhook_url, **options):
         """
         Convenience method for running bots in webhook mode
@@ -123,8 +155,9 @@ class Bot:
         if webhook_url:
             url = urlparse(webhook_url)
             app = self.create_webhook_app(url.path, loop)
+            host = os.environ.get('HOST', '0.0.0.0')
             port = int(os.environ.get('PORT', 0)) or url.port
-            web.run_app(app, port=port)
+            web.run_app(app, host=host, port=port)
 
     def stop_webhook(self):
         """
@@ -159,6 +192,7 @@ class Bot:
         """
         Set callback for default command that is called on unrecognized
         commands for 1-to-1 chats
+        If default_in_groups option is True, callback is called in groups too
 
         :Example:
 
@@ -169,6 +203,12 @@ class Bot:
         self._default = callback
         return callback
 
+    def add_inline(self, regexp, fn):
+        """
+        Manually register regexp based callback
+        """
+        self._inlines.append((regexp, fn))
+
     def inline(self, callback):
         """
         Set callback for inline queries
@@ -178,11 +218,31 @@ class Bot:
         >>> @bot.inline
         >>> def echo(iq):
         >>>     return iq.answer([
-        >>>         {"type": "text", "title": "test", "id", "0"}
+        >>>         {"type": "text", "title": "test", "id": "0"}
+        >>>     ])
+
+        >>> @bot.inline(r"myinline-(.+)")
+        >>> def echo(chat, iq, match):
+        >>>     return iq.answer([
+        >>>         {"type": "text", "title": "test", "id": "0"}
         >>>     ])
         """
-        self._inline = callback
-        return callback
+        if callable(callback):
+            self._default_inline = callback
+            return callback
+        elif isinstance(callback, str):
+            def decorator(fn):
+                self.add_inline(callback, fn)
+                return fn
+            return decorator
+        else:
+            raise TypeError('str expected {} given'.format(type(callback)))
+
+    def add_callback(self, regexp, fn):
+        """
+        Manually register regexp based callback
+        """
+        self._callbacks.append((regexp, fn))
 
     def callback(self, callback):
         """
@@ -193,9 +253,21 @@ class Bot:
         >>> @bot.callback
         >>> def echo(chat, cq):
         >>>     return cq.answer()
+
+        >>> @bot.callback(r"buttonclick-(.+)")
+        >>> def echo(chat, cq, match):
+        >>>     return chat.reply(match.group(1))
         """
-        self._callback = callback
-        return callback
+        if callable(callback):
+            self._default_callback = callback
+            return callback
+        elif isinstance(callback, str):
+            def decorator(fn):
+                self.add_callback(callback, fn)
+                return fn
+            return decorator
+        else:
+            raise TypeError('str expected {} given'.format(type(callback)))
 
     def handle(self, msg_type):
         """
@@ -236,7 +308,7 @@ class Bot:
         """
         return Chat(self, group_id, "group")
 
-    async def api_call(self, method, **params):
+    def api_call(self, method, **params):
         """
         Call Telegram API.
 
@@ -245,13 +317,18 @@ class Bot:
         :param str method: Telegram API method
         :param params: Arguments for the method call
         """
+        coro = self._api_call(method, **params)
+        # Explicitly ensure that API call is executed
+        return asyncio.ensure_future(coro)
+
+    async def _api_call(self, method, **params):
         url = "{0}/bot{1}/{2}".format(API_URL, self.api_token, method)
         logger.debug("api_call %s, %s", method, params)
 
         response = await self.session.post(url, data=params)
 
         if response.status == 200:
-            return await response.json()
+            return await response.json(loads=self.json_deserialize)
         elif response.status in RETRY_CODES:
             logger.info("Server returned %d, retrying in %d sec.",
                         response.status, RETRY_TIMEOUT)
@@ -260,11 +337,11 @@ class Bot:
             return await self.api_call(method, **params)
         else:
             if response.headers['content-type'] == 'application/json':
-                err_msg = (await response.json())["description"]
+                err_msg = (await response.json(loads=self.json_deserialize))["description"]
             else:
                 err_msg = await response.read()
             logger.error(err_msg)
-            raise RuntimeError(err_msg)
+            raise BotApiError(err_msg, response=response)
 
     async def get_me(self):
         """
@@ -349,8 +426,6 @@ class Bot:
         url = "{0}/file/bot{1}/{2}".format(API_URL, self.api_token, file_path)
         return self.session.get(url, headers=headers)
 
-    _get_user_profile_photos = partialmethod(api_call, "getUserProfilePhotos")
-
     def get_user_profile_photos(self, user_id, **options):
         """
         Get a list of profile pictures for a user
@@ -359,7 +434,8 @@ class Bot:
         :param options: Additional getUserProfilePhotos options (see
             https://core.telegram.org/bots/api#getuserprofilephotos)
         """
-        return self._get_user_profile_photos(
+        return self.api_call(
+            "getUserProfilePhotos",
             user_id=str(user_id),
             **options
         )
@@ -385,7 +461,7 @@ class Bot:
         >>> app = web.Application()
         >>> app.router.add_route('/webhook')
         """
-        update = await request.json()
+        update = await request.json(loads=self.json_deserialize)
         self._process_update(update)
         return web.Response()
 
@@ -407,6 +483,25 @@ class Bot:
             **options
         )
 
+    def delete_webhook(self):
+        '''
+        Tell Telegram to switch back to getUpdates mode
+        '''
+        return self.api_call('deleteWebhook')
+
+    @property
+    def session(self):
+        if not self._session:
+            self._session = aiohttp.ClientSession(json_serialize=self.json_serialize)
+        return self._session
+
+    def __del__(self):
+        try:
+            if self._session:
+                self._session.close()
+        except Exception as e:
+            logger.debug(e)
+
     async def _track(self, message, name):
         response = await self.session.post(
             BOTAN_URL,
@@ -415,7 +510,7 @@ class Bot:
                 "uid": message["from"]["id"],
                 "name": name
             },
-            data=json.dumps(message),
+            data=self.json_serialize(message),
             headers={'content-type': 'application/json'}
         )
         if response.status != 200:
@@ -440,18 +535,30 @@ class Bot:
                 return handler(chat, m)
 
         # No match, run default if it's a 1to1 chat
-        if not chat.is_group():
+        # However, if default_in_groups option is active, run default in any chat (not only 1to1)
+        if not chat.is_group() or self.default_in_groups:
             self.track(message, "default")
             return self._default(chat, message)
 
     def _process_inline_query(self, query):
         iq = InlineQuery(self, query)
-        return self._inline(iq)
+
+        for patterns, handler in self._inlines:
+            match = re.search(patterns, query['query'], re.I)
+            if match:
+                return handler(iq, match)
+        return self._default_inline(iq)
 
     def _process_callback_query(self, query):
         chat = Chat.from_message(self, query["message"])
         cq = CallbackQuery(self, query)
-        return self._callback(chat, cq)
+        for patterns, handler in self._callbacks:
+            match = re.search(patterns, cq.data, re.I)
+            if match:
+                return handler(chat, cq, match)
+
+        if not chat.is_group() or self.default_in_groups:
+            return self._default_callback(chat, cq)
 
     def _process_updates(self, updates):
         if not updates["ok"]:
@@ -463,15 +570,22 @@ class Bot:
 
     def _process_update(self, update):
         logger.debug("update %s", update)
+
+        # Update offset
         self._offset = max(self._offset, update["update_id"])
+
         coro = None
 
-        if "message" in update:
-            coro = self._process_message(update["message"])
-        elif "inline_query" in update:
-            coro = self._process_inline_query(update["inline_query"])
-        elif "callback_query" in update:
-            coro = self._process_callback_query(update["callback_query"])
+        # Determine update type starting with message updates
+        for ut in MESSAGE_UPDATES:
+            if ut in update:
+                coro = self._process_message(update[ut])
+                break
+        else:
+            if "inline_query" in update:
+                coro = self._process_inline_query(update["inline_query"])
+            elif "callback_query" in update:
+                coro = self._process_callback_query(update["callback_query"])
 
         if coro:
             asyncio.ensure_future(coro)
@@ -499,7 +613,7 @@ class InlineQuery:
         return self.bot.api_call(
             "answerInlineQuery",
             inline_query_id=self.query_id,
-            results=json.dumps(results),
+            results=self.bot.json_serialize(results),
             **options
         )
 
@@ -515,6 +629,7 @@ class CallbackQuery:
         self.bot = bot
         self.query_id = src['id']
         self.data = src['data']
+        self.src = src
 
     def answer(self, **options):
         return self.bot.api_call(
@@ -522,3 +637,9 @@ class CallbackQuery:
             callback_query_id=self.query_id,
             **options
         )
+
+
+class BotApiError(RuntimeError):
+    def __init__(self, *args, response):
+        super().__init__(*args)
+        self.response = response
