@@ -1,6 +1,4 @@
-import json
 import logging
-from functools import partialmethod
 
 
 logger = logging.getLogger("aiotg")
@@ -21,7 +19,7 @@ class Chat:
         """
         return self.bot.send_message(self.id, text, **options)
 
-    def reply(self, text, markup={}, parse_mode=None):
+    def reply(self, text, markup=None, parse_mode=None):
         """
         Reply to the message this `Chat` object is based on.
 
@@ -30,15 +28,18 @@ class Chat:
         :param str parse_mode: Text parsing mode (``"Markdown"``, ``"HTML"`` or
             ``None``)
         """
+        if markup is None:
+            markup = {}
+
         return self.send_text(
             text,
             reply_to_message_id=self.message["message_id"],
             disable_web_page_preview='true',
-            reply_markup=json.dumps(markup),
+            reply_markup=self.bot.json_serialize(markup),
             parse_mode=parse_mode
         )
 
-    def edit_text(self, message_id, text, markup={}, parse_mode=None):
+    def edit_text(self, message_id, text, markup=None, parse_mode=None):
         """
         Edit the message in this chat.
 
@@ -48,11 +49,14 @@ class Chat:
         :param str parse_mode: Text parsing mode (``"Markdown"``, ``"HTML"`` or
             ``None``)
         """
+        if markup is None:
+            markup = {}
+
         return self.bot.edit_message_text(
             self.id,
             message_id,
             text,
-            reply_markup=json.dumps(markup),
+            reply_markup=self.bot.json_serialize(markup),
             parse_mode=parse_mode
         )
 
@@ -66,14 +70,7 @@ class Chat:
         return self.bot.edit_message_reply_markup(
             self.id,
             message_id,
-            reply_markup=json.dumps(markup)
-        )
-
-    def _send_to_chat(self, method, **options):
-        return self.bot.api_call(
-            method,
-            chat_id=str(self.id),
-            **options
+            reply_markup=self.bot.json_serialize(markup)
         )
 
     def get_chat(self):
@@ -115,7 +112,20 @@ class Chat:
             user_id=str(user_id)
         )
 
-    send_sticker = partialmethod(_send_to_chat, "sendSticker")
+    def send_sticker(self, sticker, **options):
+        """
+        Send a sticker to the chat.
+
+        :param sticker: Sticker to send (file or string)
+        :param options: Additional sendSticker options (see
+            https://core.telegram.org/bots/api#sendsticker)
+        """
+        return self.bot.api_call(
+            "sendSticker",
+            chat_id=str(self.id),
+            sticker=sticker,
+            **options
+        )
 
     def send_audio(self, audio, **options):
         """
@@ -218,7 +228,7 @@ class Chat:
         """
         return self.bot.api_call(
             "sendVoice",
-            chat_id=self.id,
+            chat_id=str(self.id),
             voice=voice,
             **options
         )
@@ -346,12 +356,15 @@ class Chat:
 
         :return: ``True`` if this chat is a group, ``False`` otherwise
         """
-        return self.type == "group"
+        return self.type == "group" or self.type == "supergroup"
 
     def __init__(self, bot, chat_id, chat_type="private", src_message=None):
         self.bot = bot
         self.message = src_message
-        sender = src_message['from'] if src_message else {"first_name": "N/A"}
+        if src_message and 'from' in src_message:
+            sender = src_message['from']
+        else:
+            sender = {"first_name": "N/A"}
         self.sender = Sender(sender)
         self.id = chat_id
         self.type = chat_type
